local entity_forms = require 'stonehearth.lib.entity_forms.entity_forms_lib'

ethereal_storage = {}

local monkey_patches = {
   es_drop_carrying_in_storage_adjacent_action = 'stonehearth.ai.actions.drop_carrying_in_storage_adjacent_action',
   es_entities = 'radiant.modules.entities',
   es_loot_drops_component = 'stonehearth.components.loot_drops.loot_drops_component',
   es_population_faction = 'stonehearth.services.server.population.population_faction',
   es_storage_component = 'stonehearth.components.storage.storage_component',
   es_universal_storage = 'stonehearth_ace.services.server.universal_storage.universal_storage',
}

local function monkey_patch()
   for from, into in pairs(monkey_patches) do
      radiant.log.write_('ethereal_storage', 0, 'Ethereal Storage server monkey-patching sources \'' .. from .. '\' => \'' .. into .. '\'')
      local monkey_see = require('monkey_patches.' .. from)
      local monkey_do = radiant.mods.require(into)
      radiant.mixin(monkey_do, monkey_see)
   end
end

ethereal_storage.is_entity_last_access_node = function(entity, universal_storage)
   local root = entity_forms.get_root_entity(entity)
   if root and universal_storage == stonehearth_ace.universal_storage:get_storage_from_access_node(root) then
      local destination = universal_storage:get_component('destination')
      if destination and destination:get_region():get():empty() then
         return true
      end
   end

   return false
end

ethereal_storage._on_required_loaded = function()
	ethereal_storage._sv = ethereal_storage.__saved_variables:get_data()

	monkey_patch()

   local try_add_item_to_us = function(entity, destroy_if_not_added)
      if stonehearth.catalog:is_item(entity:get_uri()) then
         --radiant.log.write('ethereal_storage', 0, 'item added to ground: ' .. tostring(entity))
         local universal_storage = stonehearth_ace.universal_storage:get_universal_storage(entity:get_player_id())
         if universal_storage then
            -- first check if this item is otherwise an access point for the storage and no access points actually remain in the world
            if ethereal_storage.is_entity_last_access_node(entity, universal_storage) then
               return false
            end

            -- immediately remove the item from the world so we can try to add it to storage
            -- either it goes in storage or it gets deleted, so we don't need to save its location
            radiant.terrain.remove_entity(entity)

            if not universal_storage:add_component('stonehearth:storage'):add_item(entity) then
               -- item couldn't be added, so destroy it! no litter!
               if destroy_if_not_added then
                  radiant.entities.destroy_entity(entity)
               end
            else
               return true
            end
         end
      end
   end

   local ec = radiant.entities.get_root_entity():get_component('entity_container')
   local add_all_ground_items = function()
      local count = 0
      for id, entity in ec:each_child() do
         if try_add_item_to_us(entity) then
            count = count + 1
         end
      end

      -- then destroy all stockpiles (their items should've already been removed)
      for player_id, _ in pairs(stonehearth.player:get_players()) do
         local inventory = stonehearth.inventory:get_inventory(player_id)
         if inventory then
            local stockpiles = {}
            for id, stockpile in pairs(inventory:get_all_stockpiles()) do
               table.insert(stockpiles, stockpile)
            end
            for _, stockpile in ipairs(stockpiles) do
               radiant.entities.destroy_entity(stockpile)
            end
         end
      end

      return count
   end

   local on_game_start = function()
      ethereal_storage.game_has_started = true
      radiant.log.write('ethereal_storage', 0, 'tracing world dropped items to add to ethereal storage')
      -- This trace needs to be synchronous, so that it fires before the BFS pathfinder
      -- trace fires
      ethereal_storage._trace = ec:trace_children('tracking items on ground', _radiant.dm.TraceCategories.SYNC_TRACE)
                                 :on_added(function (id, entity)
                                    if try_add_item_to_us(entity, true) then
                                       radiant.log.write('ethereal_storage', 0, 'dropped item added to universal storage: ' .. tostring(entity))
                                    end
                                 end)

      radiant.log.write('ethereal_storage', 0, 'game started, destroying stockpiles and adding all ground items')   
      ethereal_storage._game_loaded_listener = nil
      ethereal_storage._new_game_starting_listener = nil
      -- on load, put all items on the ground into universal storage if possible
      local count = add_all_ground_items()

      radiant.log.write('ethereal_storage', 0, 'added ' .. count .. ' items to ethereal storage on load')

      ethereal_storage._storage_created_listener = radiant.events.listen(stonehearth_ace.universal_storage, 'stonehearth_ace:universal_storage:entity_created', function(args)
            radiant.log.write('ethereal_storage', 0, 'stonehearth_ace:universal_storage:entity_created, considering adding ground items... ')
            -- only add ground items if the game had already finished loading
            -- (and the universal storage hadn't been created at that time, so items weren't added to it yet)
            if args.category == stonehearth_ace.universal_storage:get_default_category() then
               radiant.log.write('ethereal_storage', 0, 'stonehearth_ace:universal_storage:entity_created, destroying stockpiles and adding all ground items')
               -- on load, put all items on the ground into universal storage if possible
               local count = add_all_ground_items()

               radiant.log.write('ethereal_storage', 0, 'added ' .. count .. ' ground items to ethereal storage on first storage entity created')
            end
         end)
   end

   ethereal_storage._game_loaded_listener = radiant.events.listen_once(radiant, 'radiant:game_loaded', on_game_start)
   ethereal_storage._new_game_starting_listener = radiant.events.listen_once(stonehearth.game_creation, 'stonehearth:world_generation_complete', on_game_start)
   
   local player_settings_set = {}
   radiant.events.listen(radiant, 'stonehearth_ace:client_gameplay_settings_set', function(player_id)
      if not player_settings_set[player_id] then
         player_settings_set[player_id] = true
         local universal_storage = stonehearth_ace.universal_storage:get_universal_storage_controller(player_id)
         if universal_storage then
            --radiant.log.write('ethereal_storage', 0, 'access node effect for '..player_id..' was '..tostring(universal_storage:get_access_node_effect())..'...')
            universal_storage:set_access_node_effect(stonehearth.client_state:get_client_gameplay_setting(player_id, 'ethereal_storage', 'universal_storage_access_node_effect', ''))
            --radiant.log.write('ethereal_storage', 0, 'access node effect for '..player_id..' being set to '..tostring(universal_storage:get_access_node_effect()))
         end
      end
   end)
end

radiant.events.listen_once(radiant, 'stonehearth_ace:server:required_loaded', ethereal_storage._on_required_loaded)

return ethereal_storage